@@ -28,6 +28,7 @@
 #include <unistd.h>
 #include <string.h>
 #include <errno.h>
+#include <math.h>
 
 #include <locale.h>
 
@@ -37,6 +38,7 @@
 #include <gdk/gdk.h>
 #include <gdk/gdkx.h>
 #include <gtk/gtk.h>
+#include <gio/gdesktopappinfo.h>
 
 #include "gnome-settings-profile.h"
 #include "gsd-marshal.h"
@@ -54,8 +56,6 @@
 #include "gvc-mixer-control.h"
 #endif /* HAVE_PULSE */
 
-#include <libupower-glib/upower.h>
-
 #define GSD_DBUS_PATH "/org/gnome/SettingsDaemon"
 #define GSD_DBUS_NAME "org.gnome.SettingsDaemon"
 #define GSD_MEDIA_KEYS_DBUS_PATH GSD_DBUS_PATH "/MediaKeys"
@@ -88,10 +88,7 @@ static const gchar introspection_xml[] =
 #define HIGH_CONTRAST "HighContrast"
 
 #define VOLUME_STEP 6           /* percents for one volume button press */
-/* FIXME: Remove when PA 0.9.23 is used */
-#ifndef PA_VOLUME_UI_MAX
-#define PA_VOLUME_UI_MAX pa_sw_volume_from_dB(+11.0)
-#endif
+#define MAX_VOLUME 65536.0
 
 #define GSD_MEDIA_KEYS_MANAGER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GSD_TYPE_MEDIA_KEYS_MANAGER, GsdMediaKeysManagerPrivate))
 
@@ -119,7 +116,7 @@ struct GsdMediaKeysManagerPrivate
 
         /* Power stuff */
         GSettings       *power_settings;
-        UpClient        *up_client;
+        GDBusProxy      *upower_proxy;
         GDBusProxy      *power_screen_proxy;
         GDBusProxy      *power_keyboard_proxy;
 
@@ -131,6 +128,7 @@ struct GsdMediaKeysManagerPrivate
 
         GDBusNodeInfo   *introspection_data;
         GDBusConnection *connection;
+        GCancellable    *bus_cancellable;
         GDBusProxy      *xrandr_proxy;
         GCancellable    *cancellable;
 
@@ -471,42 +469,54 @@ dialog_show (GsdMediaKeysManager *manager)
         gdk_display_sync (gdk_screen_get_display (manager->priv->current_screen));
 }
 
+static void
+launch_app (GAppInfo *app_info,
+	    gint64    timestamp)
+{
+	GError *error = NULL;
+        GdkAppLaunchContext *launch_context;
+
+        /* setup the launch context so the startup notification is correct */
+        launch_context = gdk_display_get_app_launch_context (gdk_display_get_default ());
+        gdk_app_launch_context_set_timestamp (launch_context, timestamp);
+
+	if (!g_app_info_launch (app_info, NULL, G_APP_LAUNCH_CONTEXT (launch_context), &error)) {
+		g_warning ("Could not launch '%s': %s",
+			   g_app_info_get_commandline (app_info),
+			   error->message);
+		g_error_free (error);
+	}
+        g_object_unref (launch_context);
+}
+
 static void
 do_url_action (GsdMediaKeysManager *manager,
-               const char          *scheme)
+               const char          *scheme,
+               gint64               timestamp)
 {
         GAppInfo *app_info;
 
         app_info = g_app_info_get_default_for_uri_scheme (scheme);
         if (app_info != NULL) {
-                GError *error = NULL;
-
-                if (!g_app_info_launch (app_info, NULL, NULL, &error)) {
-                        g_warning ("Could not launch '%s': %s",
-                                   g_app_info_get_commandline (app_info),
-                                   error->message);
-                        g_error_free (error);
-                }
+                launch_app (app_info, timestamp);
+                g_object_unref (app_info);
         } else {
                 g_warning ("Could not find default application for '%s' scheme", scheme);
-        }
+	}
 }
 
 static void
-do_media_action (GsdMediaKeysManager *manager)
+do_media_action (GsdMediaKeysManager *manager,
+		 gint64               timestamp)
 {
         GAppInfo *app_info;
 
-        app_info = g_app_info_get_default_for_type ("audio/ogg", FALSE);
+        app_info = g_app_info_get_default_for_type ("audio/x-vorbis+ogg", FALSE);
         if (app_info != NULL) {
-                GError *error = NULL;
-
-                if (!g_app_info_launch (app_info, NULL, NULL, &error)) {
-                        g_warning ("Could not launch '%s': %s",
-                                   g_app_info_get_commandline (app_info),
-                                   error->message);
-                        g_error_free (error);
-                }
+                launch_app (app_info, timestamp);
+                g_object_unref (app_info);
+        } else {
+                g_warning ("Could not find default application for '%s' mime-type", "audio/x-vorbis+ogg");
         }
 }
 
@@ -587,12 +597,47 @@ do_eject_action (GsdMediaKeysManager *manager)
         g_object_unref (volume_monitor);
 }
 
+static void
+do_home_key_action (GsdMediaKeysManager *manager,
+		    gint64               timestamp)
+{
+	GFile *file;
+	GError *error = NULL;
+	char *uri;
+
+	file = g_file_new_for_path (g_get_home_dir ());
+	uri = g_file_get_uri (file);
+	g_object_unref (file);
+
+	if (gtk_show_uri (NULL, uri, timestamp, &error) == FALSE) {
+		g_warning ("Failed to launch '%s': %s", uri, error->message);
+		g_error_free (error);
+	}
+	g_free (uri);
+}
+
+static void
+do_execute_desktop (GsdMediaKeysManager *manager,
+		    const char          *desktop,
+		    gint64               timestamp)
+{
+        GDesktopAppInfo *app_info;
+
+        app_info = g_desktop_app_info_new (desktop);
+        if (app_info != NULL) {
+                launch_app (G_APP_INFO (app_info), timestamp);
+                g_object_unref (app_info);
+        } else {
+                g_warning ("Could not find application '%s'", desktop);
+	}
+}
+
 static void
 do_touchpad_osd_action (GsdMediaKeysManager *manager, gboolean state)
 {
         dialog_init (manager);
         gsd_media_keys_window_set_action_custom (GSD_MEDIA_KEYS_WINDOW (manager->priv->dialog),
-                                                 state ? "touchpad-enabled" : "touchpad-disabled",
+                                                 state ? "input-touchpad-symbolic" : "touchpad-disabled-symbolic",
                                                  FALSE);
         dialog_show (manager);
 }
@@ -621,13 +666,16 @@ do_touchpad_action (GsdMediaKeysManager *manager)
 static void
 update_dialog (GsdMediaKeysManager *manager,
                guint vol,
-               guint max_vol,
                gboolean muted,
                gboolean sound_changed,
                gboolean quiet)
 {
-        vol = (int) (100 * (double) vol / (double) max_vol);
-        vol = CLAMP (vol, 0, 100);
+        if (!muted) {
+                vol = (int) (100 * (double) vol / PA_VOLUME_NORM);
+                vol = CLAMP (vol, 0, 100);
+        } else {
+                vol = 0.0;
+        }
 
         dialog_init (manager);
         gsd_media_keys_window_set_volume_muted (GSD_MEDIA_KEYS_WINDOW (manager->priv->dialog),
@@ -651,14 +699,13 @@ do_sound_action (GsdMediaKeysManager *manager,
                  gboolean             quiet)
 {
         gboolean old_muted, new_muted;
-        guint old_vol, new_vol, max_vol, norm_vol_step;
+        guint old_vol, new_vol, norm_vol_step;
         gboolean sound_changed;
 
         if (manager->priv->stream == NULL)
                 return;
 
-        max_vol = gvc_mixer_stream_get_can_decibel (manager->priv->stream) ? PA_VOLUME_UI_MAX : PA_VOLUME_NORM;
-        norm_vol_step = max_vol * VOLUME_STEP / 100;
+        norm_vol_step = PA_VOLUME_NORM * VOLUME_STEP / 100;
 
         /* FIXME: this is racy */
         new_vol = old_vol = gvc_mixer_stream_get_volume (manager->priv->stream);
@@ -681,7 +728,7 @@ do_sound_action (GsdMediaKeysManager *manager,
                 new_muted = FALSE;
                 /* When coming out of mute only increase the volume if it was 0 */
                 if (!old_muted || old_vol == 0)
-                        new_vol = MIN (old_vol + norm_vol_step, max_vol);
+                        new_vol = MIN (old_vol + norm_vol_step, MAX_VOLUME);
                 break;
         }
 
@@ -697,7 +744,7 @@ do_sound_action (GsdMediaKeysManager *manager,
                 }
         }
 
-        update_dialog (manager, new_vol, max_vol, new_muted, sound_changed, quiet);
+        update_dialog (manager, new_vol, new_muted, sound_changed, quiet);
 }
 
 static void
@@ -1213,44 +1260,59 @@ gnome_session_shutdown (void)
         g_object_unref (proxy);
 }
 
+static void
+upower_sleep_cb (GObject *source_object,
+                 GAsyncResult *res,
+                 gpointer user_data)
+{
+        GVariant *result;
+        GError *error = NULL;
+
+        result = g_dbus_proxy_call_finish (G_DBUS_PROXY (source_object),
+                                           res,
+                                           &error);
+        if (result == NULL) {
+                g_warning ("couldn't sleep using UPower: %s",
+                           error->message);
+                g_error_free (error);
+        } else {
+                g_variant_unref (result);
+        }
+}
+
 static void
 do_config_power_action (GsdMediaKeysManager *manager,
                         const gchar *config_key)
 {
-        gboolean ret;
-        GError *error = NULL;
         GsdPowerActionType action_type;
 
         action_type = g_settings_get_enum (manager->priv->power_settings,
                                            config_key);
         switch (action_type) {
         case GSD_POWER_ACTION_SUSPEND:
-                ret = up_client_suspend_sync (manager->priv->up_client,
-                                              NULL, &error);
-                if (!ret) {
-                        g_warning ("failed to suspend: %s",
-                                   error->message);
-                        g_error_free (error);
-                }
+                g_dbus_proxy_call (manager->priv->upower_proxy,
+                                   "Suspend",
+                                   NULL,
+                                   G_DBUS_CALL_FLAGS_NONE,
+                                   -1, NULL,
+                                   upower_sleep_cb, NULL);
                 break;
         case GSD_POWER_ACTION_INTERACTIVE:
+        case GSD_POWER_ACTION_SHUTDOWN:
                 gnome_session_shutdown ();
                 break;
         case GSD_POWER_ACTION_HIBERNATE:
-                ret = up_client_hibernate_sync (manager->priv->up_client,
-                                                NULL, &error);
-                if (!ret) {
-                        g_warning ("failed to suspend: %s",
-                                   error->message);
-                        g_error_free (error);
-                }
+                g_dbus_proxy_call (manager->priv->upower_proxy,
+                                   "Hibernate",
+                                   NULL,
+                                   G_DBUS_CALL_FLAGS_NONE,
+                                   -1, NULL,
+                                   upower_sleep_cb, NULL);
                 break;
-        case GSD_POWER_ACTION_SHUTDOWN:
         case GSD_POWER_ACTION_BLANK:
         case GSD_POWER_ACTION_NOTHING:
                 /* these actions cannot be handled by media-keys and
                  * are not used in this context */
-                g_assert_not_reached ();
                 break;
         }
 }
@@ -1350,7 +1412,7 @@ do_keyboard_brightness_action (GsdMediaKeysManager *manager,
 
         /* call into the power plugin */
         g_dbus_proxy_call (manager->priv->power_keyboard_proxy,
-                           type == SCREEN_BRIGHTNESS_UP_KEY ? "StepUp" : "StepDown",
+                           type == KEYBOARD_BRIGHTNESS_UP_KEY ? "StepUp" : "StepDown",
                            NULL,
                            G_DBUS_CALL_FLAGS_NONE,
                            -1,
@@ -1359,51 +1421,12 @@ do_keyboard_brightness_action (GsdMediaKeysManager *manager,
                            manager);
 }
 
-static void
-exec_battery_info (GsdMediaKeysManager *manager)
-{
-        gboolean ret;
-        GError *error = NULL;
-        GAppInfo *app_info;
-        GdkAppLaunchContext *launch_context;
-
-        /* setup the launch context so the startup notification is correct */
-        launch_context = gdk_display_get_app_launch_context (gdk_display_get_default ());
-        app_info = g_app_info_create_from_commandline (BINDIR "/gnome-power-statistics",
-                                                       "gnome-power-statistics",
-                                                       G_APP_INFO_CREATE_SUPPORTS_STARTUP_NOTIFICATION,
-                                                       &error);
-        if (app_info == NULL) {
-                g_warning ("failed to create application info: %s",
-                           error->message);
-                g_error_free (error);
-                goto out;
-        }
-
-        /* launch gnome-control-center */
-        ret = g_app_info_launch (app_info,
-                                 NULL,
-                                 G_APP_LAUNCH_CONTEXT (launch_context),
-                                 &error);
-        if (!ret) {
-                g_warning ("failed to launch gnome-power-statistics: %s",
-                           error->message);
-                g_error_free (error);
-                goto out;
-        }
-out:
-        g_object_unref (launch_context);
-        if (app_info != NULL)
-                g_object_unref (app_info);
-}
-
 static gboolean
 do_action (GsdMediaKeysManager *manager,
            MediaKeyType         type,
            gint64               timestamp)
 {
         char *cmd;
-        char *path;
 
         g_debug ("Launching action for key type '%d'", type);
 
@@ -1446,47 +1469,34 @@ do_action (GsdMediaKeysManager *manager,
                 do_eject_action (manager);
                 break;
         case HOME_KEY:
-                path = g_shell_quote (g_get_home_dir ());
-                cmd = g_strconcat ("nautilus --no-desktop ", path, NULL);
-                g_free (path);
-                execute (manager, cmd, FALSE, FALSE);
-                g_free (cmd);
+                do_home_key_action (manager, timestamp);
                 break;
         case SEARCH_KEY:
                 cmd = NULL;
-                if ((cmd = g_find_program_in_path ("beagle-search"))) {
-                        execute (manager, "beagle-search", FALSE, FALSE);
-                } else if ((cmd = g_find_program_in_path ("tracker-search-tool"))) {
-                        execute (manager, "tracker-search-tool", FALSE, FALSE);
-                } else {
-                        execute (manager, "gnome-search-tool", FALSE, FALSE);
-                }
+                if ((cmd = g_find_program_in_path ("tracker-search-tool")))
+                        do_execute_desktop (manager, "tracker-needle.desktop", timestamp);
+                else
+                        do_execute_desktop (manager, "gnome-search-tool.desktop", timestamp);
                 g_free (cmd);
                 break;
         case EMAIL_KEY:
-                do_url_action (manager, "mailto");
+                do_url_action (manager, "mailto", timestamp);
                 break;
         case SCREENSAVER_KEY:
         case SCREENSAVER2_KEY:
-                if ((cmd = g_find_program_in_path ("gnome-screensaver-command"))) {
-                        execute (manager, "gnome-screensaver-command --lock", FALSE, FALSE);
-                } else {
-                        execute (manager, "xscreensaver-command -lock", FALSE, FALSE);
-                }
-
-                g_free (cmd);
+                execute (manager, "gnome-screensaver-command --lock", FALSE, FALSE);
                 break;
         case HELP_KEY:
-                do_url_action (manager, "ghelp");
+                do_url_action (manager, "ghelp", timestamp);
                 break;
         case WWW_KEY:
-                do_url_action (manager, "http");
+                do_url_action (manager, "http", timestamp);
                 break;
         case MEDIA_KEY:
-                do_media_action (manager);
+                do_media_action (manager, timestamp);
                 break;
         case CALCULATOR_KEY:
-                execute (manager, "gcalctool", FALSE, FALSE);
+                do_execute_desktop (manager, "gcalctool.desktop", timestamp);
                 break;
         case PLAY_KEY:
                 return do_multimedia_player_action (manager, "Play");
@@ -1555,7 +1565,7 @@ do_action (GsdMediaKeysManager *manager,
                 do_keyboard_brightness_action (manager, type);
                 break;
         case BATTERY_KEY:
-                exec_battery_info (manager);
+                do_execute_desktop (manager, "gnome-power-statistics.desktop", timestamp);
                 break;
         case HANDLED_KEYS:
                 g_assert_not_reached ();
@@ -1665,7 +1675,6 @@ start_media_keys_idle_cb (GsdMediaKeysManager *manager)
 
         /* for the power plugin interface code */
         manager->priv->power_settings = g_settings_new (SETTINGS_POWER_DIR);
-        manager->priv->up_client = up_client_new ();
 
         /* Logic from http://git.gnome.org/browse/gnome-shell/tree/js/ui/status/accessibility.js#n163 */
         manager->priv->interface_settings = g_settings_new (SETTINGS_INTERFACE_DIR);
@@ -1757,6 +1766,12 @@ gsd_media_keys_manager_stop (GsdMediaKeysManager *manager)
 
         g_debug ("Stopping media_keys manager");
 
+        if (priv->bus_cancellable != NULL) {
+                g_cancellable_cancel (priv->bus_cancellable);
+                g_object_unref (priv->bus_cancellable);
+                priv->bus_cancellable = NULL;
+        }
+
         for (ls = priv->screens; ls != NULL; ls = ls->next) {
                 gdk_window_remove_filter (gdk_screen_get_root_window (ls->data),
                                           (GdkFilterFunc) acme_filter_events,
@@ -1783,9 +1798,9 @@ gsd_media_keys_manager_stop (GsdMediaKeysManager *manager)
                 priv->power_keyboard_proxy = NULL;
         }
 
-        if (priv->up_client) {
-                g_object_unref (priv->up_client);
-                priv->up_client = NULL;
+        if (priv->upower_proxy) {
+                g_object_unref (priv->upower_proxy);
+                priv->upower_proxy = NULL;
         }
 
         if (priv->cancellable != NULL) {
@@ -1946,6 +1961,21 @@ xrandr_ready_cb (GObject             *source_object,
         }
 }
 
+static void
+upower_ready_cb (GObject             *source_object,
+                 GAsyncResult        *res,
+                 GsdMediaKeysManager *manager)
+{
+        GError *error = NULL;
+
+        manager->priv->upower_proxy = g_dbus_proxy_new_finish (res, &error);
+        if (manager->priv->upower_proxy == NULL) {
+                g_warning ("Failed to get proxy for upower: %s",
+                           error->message);
+                g_error_free (error);
+        }
+}
+
 static void
 power_screen_ready_cb (GObject             *source_object,
                        GAsyncResult        *res,
@@ -1984,6 +2014,12 @@ on_bus_gotten (GObject             *source_object,
         GDBusConnection *connection;
         GError *error = NULL;
 
+        if (manager->priv->bus_cancellable == NULL ||
+            g_cancellable_is_cancelled (manager->priv->bus_cancellable)) {
+                g_warning ("Operation has been cancelled, so not retrieving session bus");
+                return;
+        }
+
         connection = g_bus_get_finish (res, &error);
         if (connection == NULL) {
                 g_warning ("Could not get session bus: %s", error->message);
@@ -2029,16 +2065,27 @@ on_bus_gotten (GObject             *source_object,
                           NULL,
                           (GAsyncReadyCallback) power_keyboard_ready_cb,
                           manager);
+
+        g_dbus_proxy_new (manager->priv->connection,
+                          G_DBUS_PROXY_FLAGS_NONE,
+                          NULL,
+                          "org.freedesktop.UPower",
+                          "/org/freedesktop/UPower",
+                          "org.freedesktop.UPower",
+                          NULL,
+                          (GAsyncReadyCallback) upower_ready_cb,
+                          manager);
 }
 
 static void
 register_manager (GsdMediaKeysManager *manager)
 {
         manager->priv->introspection_data = g_dbus_node_info_new_for_xml (introspection_xml, NULL);
+        manager->priv->bus_cancellable = g_cancellable_new ();
         g_assert (manager->priv->introspection_data != NULL);
 
         g_bus_get (G_BUS_TYPE_SESSION,
-                   NULL,
+                   manager->priv->bus_cancellable,
                    (GAsyncReadyCallback) on_bus_gotten,
                    manager);
 }
