@@ -24,6 +24,7 @@
 #include <libnotify/notify.h>
 #include <glib/gi18n.h>
 #include <glib/gstdio.h>
+#include <gtk/gtk.h>
 #include <cups/cups.h>
 #include <cups/ppd.h>
 
@@ -54,6 +55,22 @@ static GDBusNodeInfo *pdi_introspection_data = NULL;
 #define DBUS_TIMEOUT           60000
 #define DBUS_INSTALL_TIMEOUT 3600000
 
+#define GNOME_SESSION_DBUS_NAME                 "org.gnome.SessionManager"
+#define GNOME_SESSION_DBUS_PATH                 "/org/gnome/SessionManager"
+#define GNOME_SESSION_DBUS_IFACE                "org.gnome.SessionManager"
+#define GNOME_SESSION_CLIENT_PRIVATE_DBUS_IFACE "org.gnome.SessionManager.ClientPrivate"
+
+#define GNOME_SESSION_PRESENCE_DBUS_PATH  "/org/gnome/SessionManager/Presence"
+#define GNOME_SESSION_PRESENCE_DBUS_IFACE "org.gnome.SessionManager.Presence"
+
+enum {
+  PRESENCE_STATUS_AVAILABLE = 0,
+  PRESENCE_STATUS_INVISIBLE,
+  PRESENCE_STATUS_BUSY,
+  PRESENCE_STATUS_IDLE,
+  PRESENCE_STATUS_UNKNOWN
+};
+
 static const gchar npn_introspection_xml[] =
   "<node name='/com/redhat/NewPrinterNotification'>"
   "  <interface name='com.redhat.NewPrinterNotification'>"
@@ -81,6 +98,12 @@ static const gchar pdi_introspection_xml[] =
   "  </interface>"
   "</node>";
 
+static GMainLoop *main_loop;
+static guint      npn_registration_id;
+static guint      pdi_registration_id;
+static guint      npn_owner_id;
+static guint      pdi_owner_id;
+
 static GHashTable *
 get_missing_executables (const gchar *ppd_file_name)
 {
@@ -750,10 +773,9 @@ set_default_paper_size (const gchar *printer_name,
         GVariant    *input;
         GVariant    *output;
         GError      *error = NULL;
-        gchar      **value = NULL;
-        gchar       *paper_size;
-        char        *locale;
+        gchar       *value = NULL;
         gint         i, j, k;
+        const gchar *paper_size;
 
         proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SYSTEM,
                                                G_DBUS_PROXY_FLAGS_NONE,
@@ -770,24 +792,14 @@ set_default_paper_size (const gchar *printer_name,
                 return;
         }
 
-        /* Set default PaperSize according to the locale */
-        locale = setlocale (LC_PAPER, NULL);
-        if (!locale)
-                locale = setlocale (LC_MESSAGES, NULL);
-
-        if (!locale) {
-                g_object_unref (proxy);
-                return;
-        }
-
-        /* CLDR 2.0 alpha
-         * http://unicode.org/repos/cldr-tmp/trunk/diff/supplemental/territory_language_information.html
-         */
-        if (g_regex_match_simple ("[^_.@]{2,3}_(BZ|CA|CL|CO|CR|GT|MX|NI|PA|PH|PR|SV|US|VE)",
-                                  locale, G_REGEX_ANCHORED, G_REGEX_MATCH_ANCHORED))
-                paper_size = g_strdup ("Letter");
+        /* Set default PaperSize according to the locale
+         * FIXME: Handle more than A4 and Letter:
+         * https://bugzilla.gnome.org/show_bug.cgi?id=660769 */
+        paper_size = gtk_paper_size_get_default ();
+        if (g_str_equal (paper_size, GTK_PAPER_NAME_LETTER))
+                paper_size = "Letter";
         else
-                paper_size = g_strdup ("A4");
+                paper_size = "A4";
 
         if (ppd_file_name) {
                 ppd_file = ppdOpenFile (ppd_file_name);
@@ -801,8 +813,7 @@ set_default_paper_size (const gchar *printer_name,
                                                                                  ppd_file->groups[i].options[j].choices[k].choice,
                                                                                  strlen (paper_size)) == 0 &&
                                                             !ppd_file->groups[i].options[j].choices[k].marked) {
-                                                            value = g_new0 (gchar *, 2);
-                                                            value[0] = g_strdup (ppd_file->groups[i].options[j].choices[k].choice);
+                                                            value = g_strdup (ppd_file->groups[i].options[j].choices[k].choice);
                                                             break;
                                                         }
                                                 }
@@ -813,14 +824,15 @@ set_default_paper_size (const gchar *printer_name,
         }
 
         if (value) {
-                GVariant *array;
+                GVariantBuilder *builder;
 
-                array = g_variant_new_strv ((const gchar * const *)value, -1);
+                builder = g_variant_builder_new (G_VARIANT_TYPE ("as"));
+                g_variant_builder_add (builder, "s", value);
 
-                input = g_variant_new ("(ssv)",
+                input = g_variant_new ("(ssas)",
                                        printer_name ? printer_name : "",
-                                       "PageSize-default",
-                                       array);
+                                       "PageSize",
+                                       builder);
 
                 output = g_dbus_proxy_call_sync (proxy,
                                                  "PrinterAddOptionDefault",
@@ -837,12 +849,11 @@ set_default_paper_size (const gchar *printer_name,
                         g_error_free (error);
                 }
 
+                g_variant_builder_unref (builder);
                 g_variant_unref (input);
-                g_variant_unref (array);
-                g_strfreev (value);
+                g_free (value);
         }
 
-        g_free (paper_size);
         g_object_unref (proxy);
 }
 
@@ -1091,21 +1102,52 @@ static const GDBusInterfaceVTable interface_vtable =
   NULL
 };
 
+static void
+unregister_objects ()
+{
+        GDBusConnection *system_connection;
+        GError          *error = NULL;
+
+        system_connection = g_bus_get_sync (G_BUS_TYPE_SYSTEM, NULL, &error);
+
+        if (npn_registration_id > 0) {
+                g_dbus_connection_unregister_object (system_connection, npn_registration_id);
+                npn_registration_id = 0;
+        }
+
+        if (pdi_registration_id > 0) {
+                g_dbus_connection_unregister_object (system_connection, pdi_registration_id);
+                pdi_registration_id = 0;
+        }
+}
+
+static void
+unown_names ()
+{
+        if (npn_owner_id > 0) {
+                g_bus_unown_name (npn_owner_id);
+                npn_owner_id = 0;
+        }
+
+        if (pdi_owner_id > 0) {
+                g_bus_unown_name (pdi_owner_id);
+                pdi_owner_id = 0;
+        }
+}
+
 static void
 on_npn_bus_acquired (GDBusConnection *connection,
                      const gchar     *name,
                      gpointer         user_data)
 {
-  guint registration_id;
-
-  registration_id = g_dbus_connection_register_object (connection,
-                                                       SCP_DBUS_NPN_PATH,
-                                                       npn_introspection_data->interfaces[0],
-                                                       &interface_vtable,
-                                                       NULL,
-                                                       NULL,
-                                                       NULL);
-  g_assert (registration_id > 0);
+        npn_registration_id = g_dbus_connection_register_object (connection,
+                                                                 SCP_DBUS_NPN_PATH,
+                                                                 npn_introspection_data->interfaces[0],
+                                                                 &interface_vtable,
+                                                                 NULL,
+                                                                 NULL,
+                                                                 NULL);
+        g_assert (npn_registration_id > 0);
 }
 
 static void
@@ -1113,16 +1155,14 @@ on_pdi_bus_acquired (GDBusConnection *connection,
                      const gchar     *name,
                      gpointer         user_data)
 {
-  guint registration_id;
-
-  registration_id = g_dbus_connection_register_object (connection,
-                                                       SCP_DBUS_PDI_PATH,
-                                                       pdi_introspection_data->interfaces[0],
-                                                       &interface_vtable,
-                                                       NULL,
-                                                       NULL,
-                                                       NULL);
-  g_assert (registration_id > 0);
+        pdi_registration_id = g_dbus_connection_register_object (connection,
+                                                                 SCP_DBUS_PDI_PATH,
+                                                                 npn_introspection_data->interfaces[0],
+                                                                 &interface_vtable,
+                                                                 NULL,
+                                                                 NULL,
+                                                                 NULL);
+        g_assert (npn_registration_id > 0);
 }
 
 static void
@@ -1137,15 +1177,174 @@ on_name_lost (GDBusConnection *connection,
               const gchar     *name,
               gpointer         user_data)
 {
-  exit (1);
+        unregister_objects ();
+}
+
+static void
+session_signal_handler (GDBusConnection  *connection,
+                        const gchar      *sender_name,
+                        const gchar      *object_path,
+                        const gchar      *interface_name,
+                        const gchar      *signal_name,
+                        GVariant         *parameters,
+                        gpointer          user_data)
+{
+        guint            new_status;
+
+        g_variant_get (parameters, "(u)", &new_status);
+
+        if (new_status == PRESENCE_STATUS_IDLE ||
+            new_status == PRESENCE_STATUS_AVAILABLE) {
+                unregister_objects ();
+                unown_names ();
+
+                if (new_status == PRESENCE_STATUS_AVAILABLE) {
+                        npn_owner_id = g_bus_own_name (G_BUS_TYPE_SYSTEM,
+                                                       SCP_DBUS_NPN_NAME,
+                                                       G_BUS_NAME_OWNER_FLAGS_ALLOW_REPLACEMENT |
+                                                       G_BUS_NAME_OWNER_FLAGS_REPLACE,
+                                                       on_npn_bus_acquired,
+                                                       on_name_acquired,
+                                                       on_name_lost,
+                                                       NULL,
+                                                       NULL);
+
+                        pdi_owner_id = g_bus_own_name (G_BUS_TYPE_SYSTEM,
+                                                       SCP_DBUS_PDI_NAME,
+                                                       G_BUS_NAME_OWNER_FLAGS_ALLOW_REPLACEMENT |
+                                                       G_BUS_NAME_OWNER_FLAGS_REPLACE,
+                                                       on_pdi_bus_acquired,
+                                                       on_name_acquired,
+                                                       on_name_lost,
+                                                       NULL,
+                                                       NULL);
+                }
+        }
+}
+
+static void
+client_signal_handler (GDBusConnection  *connection,
+                       const gchar      *sender_name,
+                       const gchar      *object_path,
+                       const gchar      *interface_name,
+                       const gchar      *signal_name,
+                       GVariant         *parameters,
+                       gpointer          user_data)
+{
+        GDBusProxy *proxy;
+        GError     *error = NULL;
+        GVariant   *output;
+
+        if (g_strcmp0 (signal_name, "QueryEndSession") == 0 ||
+            g_strcmp0 (signal_name, "EndSession") == 0) {
+                proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SESSION,
+                                                       G_DBUS_PROXY_FLAGS_NONE,
+                                                       NULL,
+                                                       sender_name,
+                                                       object_path,
+                                                       interface_name,
+                                                       NULL,
+                                                       &error);
+
+                if (proxy) {
+                        output = g_dbus_proxy_call_sync (proxy,
+                                                         "EndSessionResponse",
+                                                         g_variant_new ("(bs)", TRUE, ""),
+                                                         G_DBUS_CALL_FLAGS_NONE,
+                                                         -1,
+                                                         NULL,
+                                                         &error);
+
+                        if (output) {
+                                g_variant_unref (output);
+                        }
+                        else {
+                                g_warning ("%s", error->message);
+                                g_error_free (error);
+                        }
+
+                        g_object_unref (proxy);
+                }
+                else {
+                        g_warning ("%s", error->message);
+                        g_error_free (error);
+                }
+
+                if (g_strcmp0 (signal_name, "EndSession") == 0) {
+                        g_main_loop_quit (main_loop);
+                        g_debug ("Exiting gsd-printer");
+                }
+        }
+}
+
+static gchar *
+register_gnome_session_client (const gchar *app_id,
+                               const gchar *client_startup_id)
+{
+        GDBusProxy  *proxy;
+        GVariant    *output = NULL;
+        GError      *error = NULL;
+        const gchar *client_id = NULL;
+        gchar       *result = NULL;
+
+        proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SESSION,
+                                               G_DBUS_PROXY_FLAGS_NONE,
+                                               NULL,
+                                               GNOME_SESSION_DBUS_NAME,
+                                               GNOME_SESSION_DBUS_PATH,
+                                               GNOME_SESSION_DBUS_IFACE,
+                                               NULL,
+                                               &error);
+
+        if (proxy) {
+                output = g_dbus_proxy_call_sync (proxy,
+                                                 "RegisterClient",
+                                                 g_variant_new ("(ss)", app_id, client_startup_id),
+                                                 G_DBUS_CALL_FLAGS_NONE,
+                                                 -1,
+                                                 NULL,
+                                                 &error);
+
+                if (output) {
+                        g_variant_get (output, "(o)", &client_id);
+                        if (client_id)
+                                result = g_strdup (client_id);
+                        g_variant_unref (output);
+                }
+                else {
+                        g_warning ("%s", error->message);
+                        g_error_free (error);
+                }
+
+                g_object_unref (proxy);
+        }
+        else {
+                g_warning ("%s", error->message);
+                g_error_free (error);
+        }
+
+        return result;
 }
 
 int
 main (int argc, char *argv[])
 {
-  guint npn_owner_id;
-  guint pdi_owner_id;
-  GMainLoop *loop;
+  GDBusConnection *connection;
+  gboolean         client_signal_subscription_set = FALSE;
+  GError          *error = NULL;
+  guint            client_signal_subscription_id;
+  guint            session_signal_subscription_id;
+  gchar           *object_path;
+
+  bindtextdomain (GETTEXT_PACKAGE, GNOME_SETTINGS_LOCALEDIR);
+  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
+  textdomain (GETTEXT_PACKAGE);
+  setlocale (LC_ALL, "");
+
+  npn_registration_id = 0;
+  pdi_registration_id = 0;
+  npn_owner_id = 0;
+  pdi_owner_id = 0;
 
   g_type_init ();
 
@@ -1159,29 +1358,69 @@ main (int argc, char *argv[])
           g_dbus_node_info_new_for_xml (pdi_introspection_xml, NULL);
   g_assert (pdi_introspection_data != NULL);
 
-  npn_owner_id = g_bus_own_name (G_BUS_TYPE_SYSTEM,
-                                 SCP_DBUS_NPN_NAME,
-                                 G_BUS_NAME_OWNER_FLAGS_NONE,
-                                 on_npn_bus_acquired,
-                                 on_name_acquired,
-                                 on_name_lost,
-                                 NULL,
-                                 NULL);
+  connection = g_bus_get_sync (G_BUS_TYPE_SESSION, NULL, &error);
 
-  pdi_owner_id = g_bus_own_name (G_BUS_TYPE_SYSTEM,
-                                 SCP_DBUS_PDI_NAME,
-                                 G_BUS_NAME_OWNER_FLAGS_NONE,
-                                 on_pdi_bus_acquired,
-                                 on_name_acquired,
-                                 on_name_lost,
-                                 NULL,
-                                 NULL);
+  session_signal_subscription_id =
+    g_dbus_connection_signal_subscribe (connection,
+                                        NULL,
+                                        GNOME_SESSION_PRESENCE_DBUS_IFACE,
+                                        "StatusChanged",
+                                        GNOME_SESSION_PRESENCE_DBUS_PATH,
+                                        NULL,
+                                        G_DBUS_SIGNAL_FLAGS_NONE,
+                                        session_signal_handler,
+                                        NULL,
+                                        NULL);
 
-  loop = g_main_loop_new (NULL, FALSE);
-  g_main_loop_run (loop);
+  object_path = register_gnome_session_client ("gsd-printer", "");
+  if (object_path) {
+          client_signal_subscription_id =
+                  g_dbus_connection_signal_subscribe (connection,
+                                                      NULL,
+                                                      GNOME_SESSION_CLIENT_PRIVATE_DBUS_IFACE,
+                                                      NULL,
+                                                      object_path,
+                                                      NULL,
+                                                      G_DBUS_SIGNAL_FLAGS_NONE,
+                                                      client_signal_handler,
+                                                      NULL,
+                                                      NULL);
+          client_signal_subscription_set = TRUE;
+  }
 
-  g_bus_unown_name (npn_owner_id);
-  g_bus_unown_name (pdi_owner_id);
+  if (npn_owner_id == 0)
+          npn_owner_id = g_bus_own_name (G_BUS_TYPE_SYSTEM,
+                                         SCP_DBUS_NPN_NAME,
+                                         G_BUS_NAME_OWNER_FLAGS_ALLOW_REPLACEMENT |
+                                         G_BUS_NAME_OWNER_FLAGS_REPLACE,
+                                         on_npn_bus_acquired,
+                                         on_name_acquired,
+                                         on_name_lost,
+                                         NULL,
+                                         NULL);
+
+  if (pdi_owner_id == 0)
+          pdi_owner_id = g_bus_own_name (G_BUS_TYPE_SYSTEM,
+                                         SCP_DBUS_PDI_NAME,
+                                         G_BUS_NAME_OWNER_FLAGS_ALLOW_REPLACEMENT |
+                                         G_BUS_NAME_OWNER_FLAGS_REPLACE,
+                                         on_pdi_bus_acquired,
+                                         on_name_acquired,
+                                         on_name_lost,
+                                         NULL,
+                                         NULL);
+
+  main_loop = g_main_loop_new (NULL, FALSE);
+  g_main_loop_run (main_loop);
+
+  unregister_objects ();
+  unown_names ();
+
+  if (client_signal_subscription_set)
+          g_dbus_connection_signal_unsubscribe (connection, client_signal_subscription_id);
+  g_dbus_connection_signal_unsubscribe (connection, session_signal_subscription_id);
+
+  g_free (object_path);
 
   g_dbus_node_info_unref (npn_introspection_data);
   g_dbus_node_info_unref (pdi_introspection_data);
